""" Tablib. """
try:
    # Generated by setuptools-scm.
    from ._version import version as __version__
except ImportError:
    # Some broken installation.
    __version__ = None


from tablib.core import (  # noqa: F401
    Databook,
    Dataset,
    InvalidDatasetType,
    InvalidDimensions,
    UnsupportedFormat,
    detect_format,
    import_book,
    import_set,
)
